/*
** file: maxh.cpp
**
** max Heap (binary tree on array)
*/

// Defina:
//   Em maxh.h as classes da estrutura de dados
//   Em maxh.cpp a implementação dos métodos das classes da estrutura de dados

#include <iostream>
#include <sstream>
#include "maxh.h"
using namespace std;

// Construtor: Inicializa um heap vazio com capacidade máxima definida
IMAXH::IMAXH(int nmax) : v(new int[nmax]), n(0), nv(nmax) {}

// Destruidor: Liberta a memória alocada para o vetor do heap
IMAXH::~IMAXH() {
    delete[] v;
}

// Retorna o índice do pai de um dado índice de filho
int IMAXH::getParentIndex(int childIndex) {
    return (childIndex - 1) / 2;
}

// Retorna o índice do filho à esquerda de um dado índice de pai
int IMAXH::getLeftChildIndex(int parentIndex) {
    return 2 * parentIndex + 1;
}

// Retorna o índice do filho à direita de um dado índice de pai
int IMAXH::getRightChildIndex(int parentIndex) {
    return 2 * parentIndex + 2;
}

// Método shiftUp para manter a propriedade do max heap após inserção
void IMAXH::shiftUp(int index) {
    while (index > 0 && v[getParentIndex(index)] < v[index]) {
        swap(v[getParentIndex(index)], v[index]);
        index = getParentIndex(index);
    }
}

// Método shiftDown para manter a propriedade do max heap após remoção
void IMAXH::shiftDown(int index) {
    int largest = index;
    int leftChildIndex = getLeftChildIndex(index);
    int rightChildIndex = getRightChildIndex(index);

    if (leftChildIndex < n && v[leftChildIndex] > v[largest]) {
        largest = leftChildIndex;
    }
    if (rightChildIndex < n && v[rightChildIndex] > v[largest]) {
        largest = rightChildIndex;
    }

    if (largest != index) {
        swap(v[index], v[largest]);
        shiftDown(largest);
    }
}

// Insere um novo item no heap e restaura a propriedade de heap se necessário
void IMAXH::insert(int item) {
    static bool heapFullReported = false; // Flag estática para controle da mensagem

    if (isFull()) {
        if (!heapFullReported) {
            cout << "Comando insert: Heap cheio!\n";
            heapFullReported = true; // Garante que a mensagem seja impressa apenas uma vez
        }
        return;
    }

    heapFullReported = false;       // Reset a flag se o heap não estiver cheio e a inserção continuar
    // Insere o novo elemento no final do heap
    v[n] = item;
    n++;

    // Shift up para restaurar a propriedade do heap, iniciando pelo último elemento inserido
    shiftUp(n - 1);
}

// Mostra o maior item no heap
void IMAXH::print_max() const {
    if (n > 0) {
        cout << "Max= " << v[0] << endl;
    } else {
        cout << "Comando print_max: Heap vazio!\n";
    }
}

// Mostra todos os itens no heap em formato de árvore
void IMAXH::print() const {
    if (n == 0) {
        cout << "Comando print: Heap vazio!\n";
        return;  // Encerra a função se o heap estiver vazio
    }

    cout << "Heap=\n";
    int level = 0;
    int levelNodes = 1; // Índice do primeiro elemento do próximo nível

    for (int i = 0; i < n; ++i) {
        cout << v[i];

        // Verifica se é o último elemento do nível ou do heap
        if (i + 1 == levelNodes || i == n - 1) {
            cout << "\n"; // Muda de linha
            level++;
            levelNodes = (1 << (level + 1)) - 1;
        } else {
            cout << " "; // Adiciona espaço se não for o último no nível
        }
    }
}

// Mostra o número atual de itens no heap
void IMAXH::dim() const {
    cout << "Heap tem " << n << " itens\n";
}

// Mostra a capacidade máxima do heap
void IMAXH::dim_max() const {
    cout << "Heap tem capacidade " << nv << " itens\n";
}

// Limpa o heap, removendo todos os elementos
void IMAXH::clear() {
    if (n == 0) {
        cout << "Comando clear: Heap vazio!\n";
    } else {
        n = 0;
        cout << "Heap inicializado com zero elementos\n";
    }
}

// Remove o maior elemento do heap e restaura a propriedade de heap
void IMAXH::delete_max() {
    if (n == 0) {
        cout << "Comando delete: Heap vazio!\n";
        return;
    }

    // Substitui a raiz (maior elemento) pelo último elemento no heap
    v[0] = v[n - 1];
    n--;  // Reduz o tamanho do heap

    if (n > 0) {  // Apenas se ainda há elementos no heap
        shiftDown(0);  // Reajusta a nova raiz para manter a propriedade de max heap
    }
}

// Rearranja os elementos dados para formar um heap bottom-up (~ Algoritmo de Floyd)
void IMAXH::heapify_up(const int* items, int size) {
    if (size > nv) {
        cout << "Comando heapify_up: Heap cheio!\n";
        return;
    }
    
    // Primeiro, copia todos os elementos para o vetor do heap
    for (int i = 0; i < size; ++i) {
        v[i] = items[i];
    }
    n = size; // Atualiza o contador de elementos

    // Aplica o algoritmo bottom-up a partir do último nó não folha
    for (int i = (n / 2) - 1; i >= 0; i--) {
        shiftDown(i);
    }
}

// Redimensiona o heap para uma nova capacidade máxima
int IMAXH::redim_max(int newCapacity) {
    if (newCapacity < n) {
        cout << "Comando redim_max: Heap cheio!\n";
        return nv;  // Retorna a capacidade atual, não alterada
    }

    int* newV = new int[newCapacity];
    for (int i = 0; i < n; ++i) {
        newV[i] = v[i];
    }
    delete[] v;
    v = newV;
    nv = newCapacity;
    return nv;  // Retorna a nova capacidade
}

// EOF